library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity DAQ is
    port
    (
		Input_CLK : in std_logic;
		Input_nRST : in std_logic;    
        
      --LVDS signals
      TRG_input : in std_logic;
      RDY_output: out std_logic;
      FULL_output : out std_logic;
      --connected ADC
      ADC_clk : out std_logic;
      ADC_shdn: out std_logic;
      ADC_nOE : out std_logic;
      ADC_data: in std_logic_vector(9 downto 0);
      ADC_OF  : in std_logic;
      --connect FTDI
      UART_Tx_out : out std_logic;
      UART_nRST : out std_logic;

		--USER interface
		LEDs : out std_logic_vector(3 downto 0);
		BTNs : in std_logic_vector(1 downto 0);
		  
		--Debug signal
		Debug_TRG : out std_logic;
		Debug_RDY : out std_logic;
		Debug_DataEnable : out std_logic;
		Debug_FULL : out std_logic;
		  
		GenerationOutput : out std_logic
    );
end DAQ;





architecture arch of DAQ is

    --setting constant   
    constant NumSamplPerEvent : integer := 99;   --number of samples for one event

    --component of 16 bit FIFO
   component FIFO16
	PORT
	(
		aclr		: IN STD_LOGIC  := '0';
		data		: IN STD_LOGIC_VECTOR (15 DOWNTO 0);
		rdclk		: IN STD_LOGIC ;
		rdreq		: IN STD_LOGIC ;
		wrclk		: IN STD_LOGIC ;
		wrreq		: IN STD_LOGIC ;
		q		: OUT STD_LOGIC_VECTOR (15 DOWNTO 0);
		rdempty		: OUT STD_LOGIC ;
		wrfull		: OUT STD_LOGIC ;
		wrusedw		: OUT STD_LOGIC_VECTOR (9 DOWNTO 0)
	);
    end component;

    --component uart turn
    component uart_tx is 
    port
    (
        clk : in std_logic;
        reset : in std_logic;
        start : in std_logic;
        data : in std_logic_vector(7 downto 0);
        txd : out std_logic;
        busy : out std_logic
    );
    end component;

	--component PLL for ADC domain clock
	component PLL_ADC
	PORT
	(
		areset		: IN STD_LOGIC  := '0';
		inclk0		: IN STD_LOGIC  := '0';
		c0		: OUT STD_LOGIC ;
		locked		: OUT STD_LOGIC 
	);
	end component;
    
	 --component PLL for uart clock
   component PLL_UART
	PORT
	(
		areset		: IN STD_LOGIC  := '0';
		inclk0		: IN STD_LOGIC  := '0';
		c0		: OUT STD_LOGIC ;
		locked		: OUT STD_LOGIC 
	 );
    end component;
	 
	 --component of TestGenerator
	 component TestGenerator is
	 port
	 (
	     CLK : in std_logic;
        nRST : in std_logic;
		  GenOut : out std_logic
	 );
    end component;
	 
	 --component synchronizer
	 component Synchronizer is
    generic
    (
        Reset_value : std_logic := '0';
        Num_FlipFlop : natural := 2
    );
    port
    (
        nRST : in std_logic; --asynchronous reset - low active
        CLK : in std_logic;
        Data_In : in std_logic;
        Data_Out : out std_logic
    );
	 end component;
	


	signal Sync_RST : std_logic;
    signal ClkForADC : std_logic; -- from PLL for ADC Domain
    signal ResetADC : std_logic; -- from PLL locked (no locked PLL = ADC Domain in reset)
    signal DataFromADC : std_logic_vector(9 downto 0); --data register from adc
    signal OverflowFromADC : std_logic := '0'; --of register from adc
    signal EventData : std_logic_vector(15 downto 0); --register for store to fifo
    signal DataEnable : std_logic := '0'; 

    signal Fifo_WriteReq : std_logic;
    signal Fifo_Full: std_logic;
    signal Fifo_DataOut : std_logic_vector(15 downto 0);
    signal Fifo_Empty : std_logic;
    signal Fifo_ReadReq : std_logic;
	signal Fifo_usedword : std_logic_vector(9 downto 0);

    signal ClkForUART : std_logic; --from PLL for UART domain
    signal ResetUART : std_logic; --from PLL locked for UART domain
    signal UART_start : std_logic;
    signal UART_data : std_logic_vector(7 downto 0);
    signal UART_busy : std_logic;
	signal Uart_MUX_Data : std_logic;


	signal Ready_signal : std_logic := '0';
	signal Ready_signal_S : std_logic := '0';
	signal Ready_signal_output : std_logic := '0'; 
	signal GoFlag : std_logic := '0'; --enable collect data
	 
	signal Trigger_Flag : std_logic := '0';
	signal Trigger_Flag_S : std_logic := '0';
	signal Trigger_Flag_ACK: std_logic := '0';
	 
	signal Reset_Counter : std_logic := '0';
	signal Go_Counter : std_logic := '0';
	signal EventDataMux : std_logic := '0';
	signal Event_Modulo : std_logic := '0';
	 
	signal Fifo_Free : std_logic :='0';
	 
	signal push_btn0 : std_logic := '0';
	 
	type state_data is(s0,s1,s2,s3,s4,s5,s6);
	signal data_state_reg, data_next_state : state_data := s0; 
	 
	type state_uart is(s0,s1,s2,s3,s4,s5,s6,s7);
	signal uart_state_reg, uart_next_state : state_uart := s0;


begin

    --instance of FIFO
    inst_FIFO16 : FIFO16 port map
    (
		aclr => not ResetADC,
		data => EventData, -- input data
		rdclk => ClkForUART, -- read clock, UART domain clock
		rdreq => Fifo_ReadReq, -- request for read
		wrclk => ClkForADC, -- write clock, ADC domain clock
		wrreq => Fifo_WriteReq, --request for write
		q => Fifo_DataOut,  -- output data
		rdempty => Fifo_Empty,
		wrfull => Fifo_Full,
		wrusedw => Fifo_usedword
    );

    -- instance of UART Turn
    inst_uart_tx : uart_tx port map
    (
        clk => ClkForUART, -- UART domain clock 
        reset => ResetUART,   
        start => UART_start, 
        data => UART_data,
        txd => UART_Tx_out,
        busy => UART_busy
    );    
	 
	 --instance of PLL for ADC
	 inst_PLL_ADC : PLL_ADC port map
	 (
		areset => Sync_RST,
		inclk0 => Input_CLK, 
		c0 => ClkForADC,
		locked => ResetADC		
	 );

	 --instance of PLL for UART
    inst_PLL_UART : PLL_UART port map
    (
		areset => Sync_RST,
		inclk0 => Input_CLK, 
		c0 => ClkForUART, 
		locked => ResetUART
    );
	 
	 --synchronizer reset
	 inst_Synchronizer_Reset : Synchronizer 
    generic map
    (
        Reset_value => '1',
        Num_FlipFlop => 2
	 )
    port map
    (
        nRST => Input_nRST,
        CLK => Input_CLK,
        Data_In => '0',
        Data_Out => Sync_RST
    );
	 
	 --instance of test generator	 
	 inst_TestGenerator : TestGenerator port map
	 (
	 	  CLK => Input_CLK,
        nRST => Input_nRST,
		  GenOut => GenerationOutput
	 );
	 
	 
   ADC_clk <= ClkForADC;
   	
	FULL_output <= not Fifo_Free;
	RDY_output <= not Ready_signal_output;
	
	Debug_TRG <= Trigger_Flag_S;
	Debug_RDY <=  not Ready_signal_output;
	Debug_FULL <= not Fifo_Free;
	Debug_DataEnable <= DataEnable;
	
	 

	--------------ADC Domain Clock!!!!!--------------------------------------------------

	--adc data sample
	process(ResetADC,ClkForADC)
	
	begin
		if(ResetADC = '0') then
			DataFromADC <= (others => '0');
			OverflowFromADC <= '0';
			
		elsif(ClkForADC'event and ClkForADC = '1') then
			DataFromADC <= ADC_data;
			OverflowFromADC <= ADC_OF;
			
		end if;
	
	end process;
	
	--btn 0 sync
	inst_Synchronizer_btn0 : Synchronizer 
   generic map
   (
       Reset_value => '0',
       Num_FlipFlop => 2
	)
   port map
   (
       nRST => ResetADC,
       CLK => ClkForADC,
       Data_In => not BTNs(0),
       Data_Out => push_btn0
   );
	
	--btn0 toogle
	process(ResetADC,ClkForADC)
		variable push_btn0_next : std_logic := '0';
		variable counter : unsigned(24 downto 0) := (others => '0');
	
	begin
		if(ResetADC = '0') then
			counter := (others => '0');
			push_btn0_next := '0';
			GoFlag <= '0';
			
		elsif(ClkForADC'event and ClkForADC = '1') then
			if(counter = 500000) then --read btn state every 10ms
				counter := (others => '0');
				
				if(push_btn0_next = '0' and push_btn0 = '1') then
					GoFlag <= not GoFlag;
				end if;
				
				push_btn0_next := push_btn0;
							
			else
				counter := counter + 1;
			
			end if;	
		end if;	
	end process;
	
	
	LEDs(0) <= GoFlag;

	 process(TRG_input,Trigger_Flag_ACK)
	 begin
		
		if(Trigger_Flag_ACK = '1') then
			Trigger_Flag <= '0';
		
		elsif(TRG_input'event and TRG_input = '1') then
			Trigger_Flag <= '1';
			
		end if;
	 
	 end process;
	 
	 
	 --Sync trigger flag
	 inst_Synchronizer_Trigger : Synchronizer 
    generic map
    (
        Reset_value => '0',
        Num_FlipFlop => 2
	 )
    port map
    (
        nRST => ResetADC,
        CLK => ClkForADC,
        Data_In => Trigger_Flag,
        Data_Out => Trigger_Flag_S
    );
	  
	 --Sync ready signal from uart 
	 inst_Synchronizer_ready : Synchronizer 
    generic map
    (
        Reset_value => '0',
        Num_FlipFlop => 2
	 )
    port map
    (
        nRST => ResetADC,
        CLK => ClkForADC,
        Data_In => Ready_signal,
        Data_Out => Ready_signal_S
    );
	
	
	--------------------------------------------------------------------------------------------
	--Moore finate state machine for controll Collect data
	-------------------------------------------------------------------------------------------- 
	
	-- state memory and reset
	process(ResetADC,ClkForADC)
		begin
		if(ResetADC = '0')then
			data_state_reg <= s0;
		elsif(ClkForADC'event and ClkForADC = '1') then
			data_state_reg <= data_next_state;
		end if;		
	end process;
	
	--translation function
	process(data_state_reg,ResetADC,Trigger_Flag_S,Event_Modulo,GoFlag,Ready_signal_S,FIFO_Full)
	begin
	
		data_next_state <= data_state_reg;
		
		case data_state_reg is
			
			when s0 =>
				if(ResetADC = '0')then
					data_next_state <= s0;
				elsif(ResetADC = '1') then
					data_next_state <= s1;
				end if;
				
			when s1 =>
				if(GoFlag = '1') then 
					data_next_state <= s2;
				elsif(GoFlag = '0' and Trigger_Flag_S = '1') then
					data_next_state <= s5;
				elsif(Ready_signal_S = '1') then
					data_next_state <= s5;
				end if;
					
			when s2 =>
				if(GoFlag = '0') then
					data_next_state <= s1;
				elsif(Trigger_Flag_S = '1') then
					data_next_state <= s3;
				elsif(Ready_signal_S = '1') then
					data_next_state <= s6;
				end if;
			
			when s3 =>
					if(FifO_Full = '1') then
						data_next_state <= s1;
					else
						data_next_state <= s4;
					end if;
		
			when s4 =>
				if(FifO_Full = '1') then
					data_next_state <= s1;
				elsif(Event_Modulo = '0') then
					data_next_state <= s4;
				elsif(Event_Modulo = '1') then
					data_next_state <= s1;
				end if;
			
			when s5 =>
				data_next_state <= s1;

			when s6 =>
				data_next_state <= s2;
					
			when others =>
				data_next_state <= data_state_reg;

		end case;
	end process;
	
	--output function
		process(data_state_reg)
		begin
			case(data_state_reg) is
				when s0 =>
					ADC_shdn      <= '0';
					ADC_nOE       <= '0';
					DataEnable	  <= '0';
					Fifo_WriteReq <= '0';
					Reset_Counter <= '1';
					Go_Counter    <= '0';
					EventDataMux  <= '0';
					Trigger_Flag_ACK <= '0';
					Ready_signal_output <= '1';
					
				when s1 =>
					ADC_shdn      <= '0';
					ADC_nOE       <= '0';
					DataEnable	  <= '0';
					Fifo_WriteReq <= '0';
					Reset_Counter <= '1';
					Go_Counter    <= '0';
					EventDataMux  <= '0';
					Trigger_Flag_ACK <= '0';
					Ready_signal_output <= '0';
				
				when s2 =>
					ADC_shdn      <= '0';
					ADC_nOE       <= '0';
					DataEnable	  <= '0';
					Fifo_WriteReq <= '0';
					Reset_Counter <= '1';
					Go_Counter    <= '0';
					EventDataMux  <= '0';
					Trigger_Flag_ACK <= '0';
					Ready_signal_output <= '0';
				
				when s3 =>
					ADC_shdn      <= '0';
					ADC_nOE       <= '0';
					DataEnable	  <= '0';
					Fifo_WriteReq <= '1';
					Reset_Counter <= '1';
					Go_Counter    <= '0';
					EventDataMux  <= '0';
					Trigger_Flag_ACK <= '1';
					Ready_signal_output <= '0';
				
				when s4 =>
					ADC_shdn      <= '0';
					ADC_nOE       <= '0';
					DataEnable	  <= '1';
					Fifo_WriteReq <= '1';
					Reset_Counter <= '0';
					Go_Counter    <= '1';
					EventDataMux  <= '1';
					Trigger_Flag_ACK <= '0';
					Ready_signal_output <= '0';
					
				when s5 =>
					ADC_shdn      <= '0';
					ADC_nOE       <= '0';
					DataEnable	  <= '0';
					Fifo_WriteReq <= '0';
					Reset_Counter <= '0';
					Go_Counter    <= '0';
					EventDataMux  <= '0';
					Trigger_Flag_ACK <= '1';
					Ready_signal_output <= '1';
					
				when s6 =>
					ADC_shdn      <= '0';
					ADC_nOE       <= '0';
					DataEnable	  <= '0';
					Fifo_WriteReq <= '0';
					Reset_Counter <= '0';
					Go_Counter    <= '0';
					EventDataMux  <= '0';
					Trigger_Flag_ACK <= '1';
					Ready_signal_output <= '1';

		
					
						
				when others => null;
			end case;
		end process;
	
	
	--fifo free generate signal, free = 1 if event 
	process(ResetADC,CLKForADC,Fifo_usedword)
	
		variable freewords : unsigned(9 downto 0) := (others => '0');
	
	begin
		if(ResetADC = '0') then
			freewords := (others => '0');
			Fifo_Free <= '0';
			
		elsif(CLKForADC'event and CLKForADC = '1') then
		
			freewords := 512 - unsigned(Fifo_usedword);
		
			if(freewords >= NumSamplPerEvent + 1) then
				Fifo_Free <= '1';
			
			else
				Fifo_Free <= '0';
		
			end if;
	 end if;
	
	end process;
	
	
	--event counter
	process(Reset_Counter,ClkForADC)
	
		variable counter : unsigned (16 downto 0) := (others => '0');
	
	begin
	
		if(Reset_Counter = '1') then
			counter := (others => '0');
			Event_Modulo <= '0';

		
		elsif(ClkForADC'event and ClkForADC = '1') then
			if(Go_Counter = '1') then
	
				if(counter = NumSamplPerEvent - 1) then
					Event_Modulo <= '1';
					
				else
					Counter := Counter + 1;
						
				end if;								
			end if;
		end if;
			
	end process;
	
	--event data to fifo multiplexer
	process(EventDataMux,OverflowFromADC,DataFromADC)
	begin
		if(EventDataMux = '0') then
			EventData <= "10000" & "00000000000";
		else
			EventData <= "00010" & OverflowFromADC & DataFromADC;
		end if;
	end process;
	
  

	 

    --------------UART Domain Clock!!!!!-------------------------------------------------                

	--------------------------------------------------------------------------------------------
	--Moore finate state machine for controll Turn_Uart
	-------------------------------------------------------------------------------------------- 
	
	--memory of state and reset
	process(ResetUART,ClkForUART)
	begin
		if(ResetUART = '0')then
			uart_state_reg <= s0;
		elsif(ClkForUART'event and ClkForUART = '1') then
			uart_state_reg <= uart_next_state;
		end if;		
	end process;
    
	--transitional function
	process(ResetUART,uart_state_reg, Fifo_Empty, UART_busy)
	begin
	
		uart_next_state <= uart_state_reg;
		
		case uart_state_reg is
			
			when s0 =>
				if(ResetUART = '0') then
					uart_next_state <= s0;
				elsif(ResetUART = '1') then
					uart_next_state <= s1;
				end if;
				
			when s1 =>
				if(Fifo_Empty = '1') then
					uart_next_state <= s1;
				elsif(Fifo_Empty = '0') then
					uart_next_state <= s2;
				end if;
			
			when s2 =>
				uart_next_state <= s3;
		
			when s3 =>
				if(UART_busy = '1') then
					uart_next_state <= s3;
				elsif(UART_busy = '0') then
					uart_next_state <= s4;
				end if;
				
			when s4 =>
				uart_next_state <= s5;
				
			when s5 =>
				if(UART_busy = '1') then
					uart_next_state <= s5;
				elsif(UART_busy = '0') then
					uart_next_state <= s6;
				end if;

			when s6 =>
				if(Fifo_Empty = '1') then
					uart_next_state <= s7;
				elsif(Fifo_Empty = '0') then
					uart_next_state <= s2;
				end if;

			when s7 =>
				uart_next_state <= s1;

			when others =>
				uart_next_state <= uart_state_reg;

		end case;
	end process;
	
	--output function
		process(uart_state_reg)
		begin
			case(uart_state_reg) is
				when s0 =>
					Fifo_readReq  <= '0';
					UART_start 	  <= '0';
					Ready_signal  <= '0';
					Uart_MUX_Data <= '0'; 
					UART_nRST     <= '0';
					
				when s1 =>
					Fifo_readReq  <= '0';
					UART_start    <= '0';
					Ready_signal  <= '0';
					Uart_MUX_Data <= '0'; 
					UART_nRST     <= '1';
					
				when s2 => 
					Fifo_readReq  <= '1';
					UART_start    <= '0';
					Ready_signal  <= '0';
					Uart_MUX_Data <= '0'; 
					UART_nRST     <= '1';
					
				when s3 =>
					Fifo_readReq  <= '0';
					UART_start    <= '0';
					Ready_signal  <= '0';
					Uart_MUX_Data <= '0'; 
					UART_nRST     <= '1';
					
				when s4 =>
					Fifo_readReq  <= '0';
					UART_start    <= '1';
					Ready_signal  <= '0';
					Uart_MUX_Data <= '0'; 
					UART_nRST     <= '1';
					
				when s5 =>
					Fifo_readReq  <= '0';
					UART_start    <= '0';
					Ready_signal  <= '0';
					Uart_MUX_Data <= '1'; 
					UART_nRST     <= '1';
					
				when s6 =>
					Fifo_readReq  <= '0';
					UART_start    <= '1';
					Ready_signal  <= '0';
					Uart_MUX_Data <= '1'; 
					UART_nRST     <= '1';
					
				when s7 =>
					Fifo_readReq  <= '0';
					UART_start    <= '0';
					Ready_signal  <= '1';
					Uart_MUX_Data <= '0'; 
					UART_nRST     <= '1';
					
				when others => null;
			end case;
		end process;
	
	
	UART_data <= Fifo_DataOut(7 downto 0) when Uart_MUX_Data = '1' else Fifo_DataOut(15 downto 8);
	
end arch ; -- arch
